@@ -3,14 +3,16 @@ import logging
 from cgi import escape
 
 from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
-from zope.app.publication.zopepublication import ZopePublication
+from zope.app.publication.zopepublication import ZopePublication, Cleanup
 from zope.publisher.browser import BrowserView
 from zope.publisher.interfaces.browser import IBrowserRequest
-from zope.component import adapts
+from zope.component import adapts, queryUtility
 from zope.interface import Interface
 from zope.security.proxy import removeSecurityProxy
+from zope.cachedescriptors.property import Lazy
 from ZODB.utils import p64, u64, tid_repr, oid_repr
 from ZODB.Connection import Connection
+from ZODB.interfaces import IDatabase
 from persistent import Persistent
 from persistent.TimeStamp import TimeStamp
 import transaction
@@ -72,7 +74,7 @@ class ZodbInfoView(BrowserView):
 
     @property
     def readonly(self):
-        jar = self.jar()
+        jar = self.jar
         return jar.isReadOnly()
 
     def __call__(self):
@@ -129,12 +131,17 @@ class ZodbInfoView(BrowserView):
         obj = None
         if 'oid' not in self.request:
             obj = self.findClosestPersistent()
+            # Sanity check: if we're running in standalone mode,
+            # self.context is a Folder in the just-created MappingStorage,
+            # which we're not interested in.
+            if obj is not None and obj._p_jar is not self.jar:
+                obj = None
         if obj is None:
             if 'oid' in self.request:
                 oid = int(self.request['oid'], 0)
             else:
                 oid = self.getRootOid()
-            obj = self.jar().get(p64(oid))
+            obj = self.jar.get(p64(oid))
         return obj
 
     def findClosestPersistent(self):
@@ -175,7 +182,7 @@ class ZodbInfoView(BrowserView):
         return self._tidToTimestamp(self.state.tid)
 
     def getRootOid(self):
-        root = self.jar().root()
+        root = self.jar.root()
         try:
             root = root[ZopePublication.root_name]
         except KeyError:
@@ -185,7 +192,13 @@ class ZodbInfoView(BrowserView):
     def locate_json(self, path):
         return simplejson.dumps(self.locate(path))
 
+    @Lazy
     def jar(self):
+        db = queryUtility(IDatabase, name='<target>')
+        if db is not None:
+            conn = db.open()
+            self.request.hold(Cleanup(conn.close))
+            return conn
         try:
             return self.request.annotations['ZODB.interfaces.IConnection']
         except KeyError:
@@ -195,15 +208,41 @@ class ZodbInfoView(BrowserView):
             return obj._p_jar
 
     def locate(self, path):
-        jar = self.jar()
-        oid = self.getRootOid()
+        not_found = object() # marker
+
+        # our current position
+        #   partial -- path of the last _persistent_ object
+        #   here -- path of the last object traversed
+        #   oid -- oid of the last _persistent_ object
+        #   obj -- last object traversed
         partial = here = '/'
-        obj = jar.get(p64(oid))
-        not_found = object()
-        for step in path.split('/'):
+        oid = self.getRootOid()
+        obj = self.jar.get(p64(oid))
+
+        steps = path.split('/')
+
+        if steps and steps[0]:
+            # 0x1234/sub/path -> start traversal at oid 0x1234
+            try:
+                oid = int(steps[0], 0)
+            except ValueError:
+                pass
+            else:
+                partial = here = hex(oid)
+                try:
+                    obj = self.jar.get(p64(oid))
+                except KeyError:
+                    oid = self.getRootOid()
+                    return dict(error='Not found: %s' % steps[0],
+                                partial_oid=oid,
+                                partial_path='/',
+                                partial_url=self.getUrl(oid))
+                steps = steps[1:]
+
+        for step in steps:
             if not step:
                 continue
-            if here != '/':
+            if not here.endswith('/'):
                 here += '/'
             here += step.encode('utf-8')
             try:
@@ -228,16 +267,13 @@ class ZodbInfoView(BrowserView):
                     url=self.getUrl(oid))
 
     def getUrl(self, oid=None, tid=None):
-        url = "@@zodbbrowser?oid="
-        if oid is not None:
-            url += str(oid)
-        else:
-            url += str(self.getObjectId())
-
+        if oid is None:
+            oid = self.getObjectId()
+        url = "@@zodbbrowser?oid=0x%x" % oid
         if tid is None and 'tid' in self.request:
             url += "&tid=" + self.request['tid']
         elif tid is not None:
-            url += "&tid=" + str(tid)
+            url += "&tid=0x%x" % tid
         return url
 
     def getBreadcrumbs(self):
@@ -252,6 +288,11 @@ class ZodbInfoView(BrowserView):
             else:
                 if breadcrumbs:
                     breadcrumbs.append(('/', None))
+                if not state.getName() and state.getParentState() is None:
+                    # not using hex() because we don't want L suffixes for
+                    # 64-bit values
+                    breadcrumbs.append(('0x%x' % state.getObjectId(), url))
+                    break
                 breadcrumbs.append((state.getName() or '???', url))
             state = state.getParentState()
             if state is None:
@@ -294,9 +335,10 @@ class ZodbInfoView(BrowserView):
         results = []
         for d in self.history:
             try:
-                state = ZodbObjectState(self.obj, d['tid'],
-                                        _history=self.history).asDict()
-                error = None
+                interp = ZodbObjectState(self.obj, d['tid'],
+                                         _history=self.history)
+                state = interp.asDict()
+                error = interp.getError()
             except Exception, e:
                 state = {}
                 error = '%s: %s' % (e.__class__.__name__, e)
