@@ -13,14 +13,18 @@ import logging
 
 from ZEO.ClientStorage import ClientStorage
 from ZODB.DB import DB
+from ZODB.MappingStorage import MappingStorage
 from ZODB.FileStorage.FileStorage import FileStorage
+from ZODB.interfaces import IDatabase
 from zope.app.server.servertype import IServerType
 from zope.app.appsetup.appsetup import SystemConfigurationParticipation
-from zope.component import getUtility
+from zope.component import getUtility, provideUtility
 from zope.server.taskthreads import ThreadedTaskDispatcher
 from zope.event import notify
 import zope.app.component.hooks
 
+from zodbbrowser.state import install_provides_hack
+
 
 class Options(object):
     listen_on = ('localhost', 8070)
@@ -60,8 +64,6 @@ class Options(object):
           <role id="zope.Member" title="Site Member" />
           <grant permission="zope.View"
                  role="zope.Anonymous" />
-          <grant permission="zope.app.dublincore.view"
-                 role="zope.Anonymous" />
           <grantAll role="zope.Manager" />
           <!-- INSECURE -->
           <grant permission="zope.ManageContent"
@@ -176,14 +178,20 @@ def main(args=None, start_serving=True):
     else:
         parser.error('please specify a database')
 
+    internal_db = DB(MappingStorage())
+
     configure(options)
 
-    notify(zope.app.appsetup.interfaces.DatabaseOpened(db))
+    provideUtility(db, IDatabase, name='<target>')
 
-    start_server(options, db)
+    notify(zope.app.appsetup.interfaces.DatabaseOpened(internal_db))
+
+    start_server(options, internal_db)
 
     notify(zope.app.appsetup.interfaces.ProcessStarting())
 
+    install_provides_hack()
+
     if start_serving:
         serve_forever()
 
