@@ -1,12 +1,18 @@
+import logging
+
 from persistent.dict import PersistentDict
 from persistent.list import PersistentList
 from persistent.mapping import PersistentMapping
 from zope.component import adapts, getMultiAdapter
 from zope.interface import implements, Interface
+from zope.interface.interfaces import IInterface
+from zope.interface.interface import InterfaceClass
 from zope.proxy import removeAllProxies
 from zope.traversing.interfaces import IContainmentRoot
 from ZODB.utils import u64
 
+import zope.interface.declarations
+
 # be compatible with Zope 3.4, but prefer the modern package structure
 try:
     from zope.container.sample import SampleContainer
@@ -24,6 +30,47 @@ except ImportError:
 from zodbbrowser.interfaces import IStateInterpreter, IObjectHistory
 
 
+log = logging.getLogger(__name__)
+
+
+real_Provides = zope.interface.declarations.Provides
+
+
+def install_provides_hack():
+    """Monkey-patch zope.interface.Provides with a more lenient version.
+
+    A common result of missing modules in sys.path is that you cannot
+    unpickle objects that have been marked with directlyProvides() to
+    implement interfaces that aren't currently available.  Those interfaces
+    are replaced by persistent broken placeholders, which aren classes,
+    not interfaces, and aren't iterable, causing TypeErrors during unpickling.
+    """
+    zope.interface.declarations.Provides = Provides
+
+
+def flatten_interfaces(args):
+    result = []
+    for a in args:
+        if isinstance(a, (list, tuple)):
+            result.extend(flatten_interfaces(a))
+        elif IInterface.providedBy(a):
+            result.append(a)
+        else:
+            log.warning('  replacing %s with a placeholder', repr(a))
+            result.append(InterfaceClass(a.__name__,
+                            __module__='broken ' + a.__module__))
+    return result
+
+
+def Provides(cls, *interfaces):
+    try:
+        return real_Provides(cls, *interfaces)
+    except TypeError, e:
+        log.warning('Suppressing TypeError while unpickling Provides: %s', e)
+        args = flatten_interfaces(interfaces)
+        return real_Provides(cls, *args)
+
+
 class ZodbObjectState(object):
     implements(IStateInterpreter)
 
@@ -36,14 +83,23 @@ class ZodbObjectState(object):
         self.history = _history
         self.tid = None
         self.requestedTid = tid
+        self.loadError = None
         self._load()
 
     def _load(self):
         self.tid = self.history.lastChange(self.requestedTid)
-        loadedState = self.history.loadState(self.tid)
-        self.state = getMultiAdapter((self.obj, loadedState,
-                                      self.requestedTid),
-                                     IStateInterpreter)
+        try:
+            loadedState = self.history.loadState(self.tid)
+        except Exception, e:
+            self.loadError = "%s: %s" % (e.__class__.__name__, e)
+            self.state = LoadErrorState(self.loadError, self.requestedTid)
+        else:
+            self.state = getMultiAdapter((self.obj, loadedState,
+                                         self.requestedTid),
+                                         IStateInterpreter)
+
+    def getError(self):
+        return self.loadError
 
     def listAttributes(self):
         return self.state.listAttributes()
@@ -59,9 +115,11 @@ class ZodbObjectState(object):
         if name is None:
             # __name__ is not in the pickled state, but it may be defined
             # via other means (e.g. class attributes, custom __getattr__ etc.)
-            name = getattr(self.obj, '__name__', None)
-        if not name:
-            name = '???'
+            try:
+                name = getattr(self.obj, '__name__', None)
+            except Exception:
+                # Ouch.  Oh well, we can't determine the name.
+                pass
         return name
 
     def asDict(self):
@@ -83,6 +141,33 @@ class ZodbObjectState(object):
             return ZodbObjectState(parent, self.requestedTid)
 
 
+class LoadErrorState(object):
+    """Placeholder for when an object's state could not be loaded"""
+    implements(IStateInterpreter)
+
+    def __init__(self, error, tid):
+        self.error = error
+        self.tid = tid
+
+    def getError(self):
+        return self.error
+
+    def getName(self):
+        return None
+
+    def getParent(self):
+        return None
+
+    def listAttributes(self):
+        return []
+
+    def listItems(self):
+        return None
+
+    def asDict(self):
+        return {}
+
+
 class GenericState(object):
     """Most persistent objects represent their state as a dict."""
     adapts(Interface, dict, None)
@@ -92,6 +177,9 @@ class GenericState(object):
         self.state = state
         self.tid = tid
 
+    def getError(self):
+        return None
+
     def getName(self):
         return self.state.get('__name__')
 
@@ -206,6 +294,9 @@ class FallbackState(object):
     def __init__(self, type, state, tid):
         self.state = state
 
+    def getError(self):
+        return None
+
     def getName(self):
         return None
 
