@@ -1,12 +1,14 @@
 import unittest
 import transaction
 import sys
+import gc
 
 from ZODB.utils import u64, p64, tid_repr, oid_repr
+from ZODB.interfaces import IDatabase
 from zope.app.container.btree import BTreeContainer
 from zope.app.container.interfaces import IContained
 from zope.app.testing import setup
-from zope.component import provideAdapter
+from zope.component import provideAdapter, getGlobalSiteManager
 from zope.interface import implements
 from zope.publisher.browser import TestRequest
 from zope.traversing.interfaces import IContainmentRoot
@@ -19,6 +21,24 @@ from zodbbrowser.testing import SimpleValueRenderer
 from realdb import RealDatabaseTest
 
 
+class DatabaseStub(object):
+    implements(IDatabase)
+
+    opened = 0
+
+    def open(self):
+        self.opened += 1
+        return ConnectionStub(self)
+
+
+class ConnectionStub(object):
+    def __init__(self, db):
+        self.db = db
+
+    def close(self):
+        self.db.opened -= 1
+
+
 class RootFolderStub(BTreeContainer):
     implements(IContainmentRoot)
 
@@ -113,9 +133,9 @@ class TestZodbInfoViewWithRealDb(RealDatabaseTest):
 
     def testGetJar(self):
         view = ZodbInfoView(self.root, TestRequest())
-        self.assertEquals(view.jar(), self.root._p_jar)
+        self.assertEquals(view.jar, self.root._p_jar)
         view = ZodbInfoView(self.root['stub']['member'], TestRequest())
-        self.assertEquals(view.jar(), self.root._p_jar)
+        self.assertEquals(view.jar, self.root._p_jar)
 
     def testSelectObjectToView_use_context(self):
         view = ZodbInfoView(self.root, TestRequest())
@@ -133,7 +153,7 @@ class TestZodbInfoViewWithRealDb(RealDatabaseTest):
 
     def testSelectObjectToView_find_parent_fail_fall_back_to_root(self):
         view = ZodbInfoView(self.root['stub']['member']['notpersistent'], TestRequest())
-        view.jar = lambda: self.root._p_jar
+        view.jar = self.root._p_jar
         self.assertEquals(view.selectObjectToView(), self.root)
 
     def testSelectObjectToView_by_oid(self):
@@ -176,27 +196,48 @@ class TestZodbInfoViewWithRealDb(RealDatabaseTest):
     def testLocate(self):
         view = self._zodbInfoView(self.root, TestRequest())
         jsonResult = view.locate_json('/')
-        self.assertTrue('"url": "@@zodbbrowser?oid=0"' in jsonResult)
+        self.assertTrue('"url": "@@zodbbrowser?oid=0x0"' in jsonResult)
         self.assertTrue('"oid": 0' in jsonResult)
         jsonResult = view.locate_json('/stub/member/notpersistent')
         self.assertTrue('"partial_url"' in jsonResult)
         self.assertTrue('"partial_oid"' in jsonResult)
         self.assertTrue('"error": "Not persistent: /stub/member/notpersistent"')
         jsonResult = view.locate_json('/stub/nonexistent')
-        self.assertTrue('"partial_url": "@@zodbbrowser?oid=1"' in jsonResult)
+        self.assertTrue('"partial_url": "@@zodbbrowser?oid=0x1"' in jsonResult)
         self.assertTrue('"partial_path": "/stub", ' in jsonResult)
         self.assertTrue('"error": "Not found: /stub/nonexistent"}' in jsonResult)
         self.assertTrue('"partial_oid": 1' in jsonResult)
 
+    def testLocateStartWithOID(self):
+        view = self._zodbInfoView(self.root, TestRequest())
+        jsonResult = view.locate_json('0')
+        self.assertTrue('"url": "@@zodbbrowser?oid=0x0"' in jsonResult)
+        self.assertTrue('"oid": 0' in jsonResult)
+        jsonResult = view.locate_json('0x0')
+        self.assertTrue('"url": "@@zodbbrowser?oid=0x0"' in jsonResult)
+        self.assertTrue('"oid": 0' in jsonResult)
+        jsonResult = view.locate_json('0x1/nonexistent')
+        self.assertTrue('"partial_url": "@@zodbbrowser?oid=0x1"' in jsonResult)
+        self.assertTrue('"partial_path": "0x1", ' in jsonResult)
+        self.assertTrue('"error": "Not found: 0x1/nonexistent"}' in jsonResult)
+        self.assertTrue('"partial_oid": 1' in jsonResult)
+
+    def testLocateStartWithOID_that_does_not_exist(self):
+        view = self._zodbInfoView(self.root, TestRequest())
+        jsonResult = view.locate_json('0x1234')
+        self.assertTrue('"partial_url": "@@zodbbrowser?oid=0x0"' in jsonResult)
+        self.assertTrue('"partial_path": "/", ' in jsonResult)
+        self.assertTrue('"error": "Not found: 0x1234"}' in jsonResult)
+        self.assertTrue('"partial_oid": 0' in jsonResult)
+
     def testGetUrl(self):
         view = self._zodbInfoView(self.root, TestRequest())
-        self.assertEquals(view.getUrl(), '@@zodbbrowser?oid=' +
-                          str(u64(self.root._p_oid)))
+        self.assertEquals(view.getUrl(), '@@zodbbrowser?oid=0x%x' %
+                          u64(self.root._p_oid))
         view = self._zodbInfoView(self.root, TestRequest())
-        self.assertEquals(view.getUrl(1, 2), '@@zodbbrowser?oid=1&tid=2')
+        self.assertEquals(view.getUrl(1, 2), '@@zodbbrowser?oid=0x1&tid=0x2')
         view = ZodbInfoView(self.root, TestRequest(form={'tid': '2'}))
-        self.assertEquals(view.getUrl(1), '@@zodbbrowser?oid=1&tid=2')
-
+        self.assertEquals(view.getUrl(1), '@@zodbbrowser?oid=0x1&tid=2')
 
 
 class ZodbObjectStateStub(object):
@@ -256,59 +297,53 @@ class TestZodbInfoViewBreadcrumbs(unittest.TestCase):
     def test_root(self):
         view = self.createView(self.root)
         self.assertEquals(view.getBreadcrumbs(),
-                          [('/', '@@zodbbrowser?oid=1'),
+                          [('/', '@@zodbbrowser?oid=0x1'),
                           ])
 
     def test_non_root(self):
         view = self.createView(self.foo)
         self.assertEquals(view.getBreadcrumbs(),
-                          [('/', '@@zodbbrowser?oid=1'),
-                           ('foo','@@zodbbrowser?oid=27'),
+                          [('/', '@@zodbbrowser?oid=0x1'),
+                           ('foo','@@zodbbrowser?oid=0x1b'),
                           ])
 
     def test_more_levels(self):
         view = self.createView(self.foobar)
         self.assertEquals(view.getBreadcrumbs(),
-                          [('/', '@@zodbbrowser?oid=1'),
-                           ('foo','@@zodbbrowser?oid=27'),
+                          [('/', '@@zodbbrowser?oid=0x1'),
+                           ('foo','@@zodbbrowser?oid=0x1b'),
                            ('/', None),
-                           ('bar','@@zodbbrowser?oid=32'),
+                           ('bar','@@zodbbrowser?oid=0x20'),
                           ])
 
     def test_unknown(self):
         view = self.createView(self.unknown)
         self.assertEquals(view.getBreadcrumbs(),
-                          [('/', '@@zodbbrowser?oid=1'),
-                           ('...', None),
-                           ('/', None),
-                           ('???','@@zodbbrowser?oid=15'),
+                          [('0xf','@@zodbbrowser?oid=0xf'),
                           ])
 
     def test_unknown_child(self):
         view = self.createView(self.unknown_child)
         self.assertEquals(view.getBreadcrumbs(),
-                          [('/', '@@zodbbrowser?oid=1'),
-                           ('...', None),
+                          [('0xf','@@zodbbrowser?oid=0xf'),
                            ('/', None),
-                           ('???','@@zodbbrowser?oid=15'),
-                           ('/', None),
-                           ('child', '@@zodbbrowser?oid=17'),
+                           ('child', '@@zodbbrowser?oid=0x11'),
                           ])
 
     def test_unparented(self):
         view = self.createView(self.unparented)
         self.assertEquals(view.getBreadcrumbs(),
-                          [('/', '@@zodbbrowser?oid=1'),
+                          [('/', '@@zodbbrowser?oid=0x1'),
                            ('...', None),
                            ('/', None),
-                           ('wat', '@@zodbbrowser?oid=19'),
+                           ('wat', '@@zodbbrowser?oid=0x13'),
                           ])
 
     def test_unnamed_direct_child_of_root(self):
         view = self.createView(self.unnamed)
         self.assertEquals(view.getBreadcrumbs(),
-                          [('/', '@@zodbbrowser?oid=1'),
-                           ('???', '@@zodbbrowser?oid=55'),
+                          [('/', '@@zodbbrowser?oid=0x1'),
+                           ('???', '@@zodbbrowser?oid=0x37'),
                           ])
 
 
@@ -318,6 +353,28 @@ class TestZodbInfoView(unittest.TestCase):
         if first != second:
             self.fail('\n%r !=\n%r' % (first, second))
 
+    def addCleanUp(self, fn, *args, **kw):
+        self.cleanups.append((fn, args, kw))
+
+    def setUp(self):
+        self.cleanups = []
+
+    def tearDown(self):
+        for fn, args, kw in reversed(self.cleanups):
+            fn(*args, **kw)
+
+    def testGetJar_uses_explicit_target_db(self):
+        stub_db = DatabaseStub()
+        registry = getGlobalSiteManager()
+        registry.registerUtility(stub_db, IDatabase, name='<target>')
+        self.addCleanUp(registry.unregisterUtility,
+                        stub_db, IDatabase, name='<target>')
+        view = ZodbInfoView(object(), TestRequest())
+        self.assertEquals(view.jar.db, stub_db)
+        del view
+        gc.collect()
+        self.assertEquals(stub_db.opened, 0)
+
     def test_getPath(self):
         view = ZodbInfoView(None, None)
         view.getBreadcrumbs = lambda: [('/', None), ('foo', None),
