@@ -1,4 +1,4 @@
-import unittest
+import unittest2 as unittest
 import sys
 
 from ZODB.utils import p64
@@ -45,6 +45,10 @@ class PersistentFrob(Persistent):
         return '<PersistentFrob>'
 
 
+class Struct(object):
+    pass
+
+
 class FrobRenderer(object):
     adapts(Frob)
     implements(IValueRenderer)
@@ -59,6 +63,17 @@ class FrobRenderer(object):
             return '<Frob>'
 
 
+class StructRenderer(object):
+    adapts(Struct)
+    implements(IValueRenderer)
+
+    def __init__(self, context):
+        self.context = context
+
+    def render(self, tid=None):
+        return '<span class="struct">Struct</span>'
+
+
 class TestGenericValue(unittest.TestCase):
 
     def test_interface_compliance(self):
@@ -100,6 +115,7 @@ class TestTupleValue(unittest.TestCase):
         setup.placelessSetUp()
         provideAdapter(GenericValue)
         provideAdapter(FrobRenderer)
+        provideAdapter(StructRenderer)
 
     def tearDown(self):
         setup.placelessTearDown()
@@ -122,6 +138,13 @@ class TestTupleValue(unittest.TestCase):
         self.assertEquals(TupleValue((Frob(), )).render(tid=42),
                           '(<Frob [tid=42]>, )')
 
+    def test_nested_structs(self):
+        self.assertEquals(TupleValue((Struct(), )).render(),
+              '(<span class="struct"><span class="struct">Struct, )</span></span>')
+        self.assertEquals(TupleValue((Struct(), Struct())).render(),
+              '(<span class="struct"><span class="struct">Struct,</span>'
+              '<br /><span class="struct">Struct)</span></span>')
+
 
 class TestListValue(unittest.TestCase):
 
@@ -129,6 +152,7 @@ class TestListValue(unittest.TestCase):
         setup.placelessSetUp()
         provideAdapter(GenericValue)
         provideAdapter(FrobRenderer)
+        provideAdapter(StructRenderer)
 
     def tearDown(self):
         setup.placelessTearDown()
@@ -151,18 +175,26 @@ class TestListValue(unittest.TestCase):
         self.assertEquals(ListValue([Frob()]).render(tid=42),
                           '[<Frob [tid=42]>]')
 
+    def test_nested_structs(self):
+        self.assertEquals(ListValue((Struct(), )).render(),
+              '[<span class="struct"><span class="struct">Struct]</span></span>')
+        self.assertEquals(ListValue((Struct(), Struct())).render(),
+              '[<span class="struct"><span class="struct">Struct,</span>'
+              '<br /><span class="struct">Struct]</span></span>')
+
+
 
 class TestDictValue(unittest.TestCase):
 
-    def assertEquals(self, first, second):
-        if first != second:
-            self.fail('\n%r !=\n%r' % (first, second))
+    maxDiff = None
 
     def setUp(self):
         setup.placelessSetUp()
         provideAdapter(GenericValue)
         provideAdapter(FrobRenderer)
+        provideAdapter(StructRenderer)
         provideAdapter(DictValue)
+        self.addTypeEqualityFunc(str, 'assertMultiLineEqual')
 
     def tearDown(self):
         setup.placelessTearDown()
@@ -186,20 +218,23 @@ class TestDictValue(unittest.TestCase):
     def test_truly_long_dicts(self):
         self.assertEquals(DictValue(
             {'some long key name': 'some long value',
-             'some other long key name': 'some other long value'}).render(),
-           "{<span class=\"dict\">'some long key name': 'some long value',"
+             'some other long key name': 'some other long value'}).render(threshold=50),
+           "{<span class=\"struct\">'some long key name': 'some long value',"
            "<br />'some other long key name': 'some other long value'}</span>")
 
     def test_nested_dicts(self):
         self.assertEquals(DictValue(
             {'A': {'some long key name': 'some long value',
-                       'some other long key name': 'some other long value'},
-             'B': ['something else entirely']}).render(),
-           "{<span class=\"dict\">'A':"
-                " {<span class=\"dict\">'some long key name':"
-                " 'some long value',"
+                   'some other long key name': 'some other long value',
+                   'struct': Struct()},
+             'B': ['something else entirely']}).render().replace('<br />', '\n<br />'),
+           "{<span class=\"struct\">'A':"
+                " {<span class=\"struct\">'some long key name':"
+                " 'some long value',\n"
                 "<br />'some other long key name':"
-                " 'some other long value'},</span>"
+                " 'some other long value',\n"
+                "<br />'struct':"
+                " <span class=\"struct\">Struct}</span>,</span>\n"
            "<br />'B': ['something else entirely'] (1 item)}</span>")
 
     def test_tid_is_preserved(self):
